@@ -485,14 +485,15 @@ fz_copy_pixmap_area_converting_seps(fz_context *ctx, fz_pixmap *src, fz_pixmap *
 						if (dname && !strcmp(name, dname))
 							goto map_device_n_spot;
 					}
-					for (j = 0; j < dseps_n; j++)
+					for (k = 0; k < dseps_n; k++)
 					{
-						const char *dname = dseps->name[j];
+						const char *dname = dseps->name[k];
+						int state = sep_state(dseps, k);
+						if (state != FZ_SEPARATION_SPOT)
+							continue;
 						if (dname && !strcmp(name, dname))
-						{
-							j += dc;
 							goto map_device_n_spot;
-						}
+						j++;
 					}
 				}
 				if (0)
@@ -1069,7 +1070,7 @@ fz_convert_separation_colors(fz_context *ctx,
 	fz_separations *dst_seps, fz_colorspace *dst_cs, float *dst_color,
 	fz_color_params color_params)
 {
-	int i, j, n, dc, ds, dn, pred;
+	int i, j, k, n, dc, ds, dn;
 	float remainders[FZ_MAX_COLORS];
 	int remaining = 0;
 
@@ -1088,7 +1089,6 @@ fz_convert_separation_colors(fz_context *ctx,
 		dst_color[i] = 0;
 
 	n = fz_colorspace_n(ctx, src_cs);
-	pred = 0;
 	for (i = 0; i < n; i++)
 	{
 		const char *name = fz_colorspace_colorant(ctx, src_cs, i);
@@ -1107,19 +1107,20 @@ fz_convert_separation_colors(fz_context *ctx,
 		if (!strcmp(name, "None"))
 			continue;
 
-		/* The most common case is that the colorant we match is the
-		 * one after the one we matched before, so optimise for that. */
-		for (j = pred; j < ds; j++)
-		{
-			const char *dname = dst_seps->name[j];
-			if (dname && !strcmp(name, dname))
-				goto found_sep;
-		}
-		for (j = 0; j < pred; j++)
+		k = 0;
+		for (j = 0; j < ds; j++)
 		{
+			fz_separation_behavior beh = fz_separation_current_behavior(ctx, dst_seps, j);
 			const char *dname = dst_seps->name[j];
 			if (dname && !strcmp(name, dname))
+			{
+				if (beh == FZ_SEPARATION_DISABLED)
+					goto found_disabled;
 				goto found_sep;
+			}
+			if (beh != FZ_SEPARATION_SPOT)
+				continue;
+			k++;
 		}
 		for (j = 0; j < dc; j++)
 		{
@@ -1129,14 +1130,19 @@ fz_convert_separation_colors(fz_context *ctx,
 		}
 		if (0) {
 found_sep:
-			dst_color[j+dc] = src_color[i];
-			pred = j+1;
+			dst_color[k+dc] = src_color[i];
 		}
 		else if (0)
 		{
 found_process:
 			dst_color[j] += src_color[i];
 		}
+		else if (0)
+		{
+found_disabled:
+			/* Don't do anything with this value */
+			{}
+		}
 		else
 		{
 			if (remaining == 0)
