@@ -449,7 +449,7 @@ pdf_repair_xref_base(fz_context *ctx, pdf_document *doc)
 	doc->repair_in_progress = 1;
 
 	pdf_drop_page_tree_internal(ctx, doc);
-	doc->page_tree_broken = 0;
+	doc->use_page_tree_map = 1;
 	pdf_forget_xref(ctx, doc);
 
 	fz_seek(ctx, doc->file, 0, 0);
